import logging
import subprocess
import json
import os

from ossbom.converters.factory import SBOMConverterFactory
from ossbom.model.ossbom import OSSBOM

logger = logging.getLogger(__name__)


def create_sbom_from_requirements(requirements_file: str) -> OSSBOM:

    try:
        # This command generates an SBOM for the active virtual environment in JSON format
        result = subprocess.run(
            ['cyclonedx-py', 'requirements', requirements_file],
            check=True,
            capture_output=True,
            text=True,
            env=os.environ.copy()
        )

        ret = result.stdout

        cyclone_dict = json.loads(ret)

        ossbom = SBOMConverterFactory.from_cyclonedx_dict(cyclone_dict)

        return ossbom

    except subprocess.CalledProcessError as e:
        logging.error(f"Error running creating SBOM: {e}")
        logging.debug(e.stderr)
        logging.debug("--")
        logging.debug(e.stdout)
        raise e


def update_sbom_from_requirements(ossbom: OSSBOM, requirements_file: str) -> OSSBOM:
    sbom = create_sbom_from_requirements(requirements_file)
    ossbom.add_components(sbom.get_components())
    
    return ossbom


def create_sbom_from_env() -> OSSBOM:

    try:
        # This command generates an SBOM for the active virtual environment in JSON format
        result = subprocess.run(
            ['cyclonedx-py', 'environment'],
            check=True,
            capture_output=True,
            text=True,
            env=os.environ.copy()
        )

        ret = result.stdout

        cyclone_dict = json.loads(ret)

        ossbom = SBOMConverterFactory.from_cyclonedx_dict(cyclone_dict)

        return ossbom

    except subprocess.CalledProcessError as e:
        logging.error(f"Error running creating SBOM: {e}")
        logging.debug(result.stderr)
        logging.debug("--")
        logging.debug(result.stdout)
        raise e



import json
import logging
import os
import re
import subprocess
from pathlib import Path

from typing import List

from ossbom.model.ossbom import OSSBOM
from ossbom.model.component import Component
from ossbom.model.dependency_env import DependencyEnv

logger = logging.getLogger(__name__)


def exec_command(command: str, cwd=None):
    try:
        # Run the yarn command with the specified cwd
        result = subprocess.run(
            command.split(" "),
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True,
            cwd=cwd if cwd else None,  # Explicitly set to None if not used
            check=True
        )
        # Parse the output
        return result.stdout
    except subprocess.CalledProcessError as e:
        logger.error("Error while running yarn:", e)
        return e.stdout


def node_modules_directory_exists(project_folder: str) -> bool:
    return os.path.isdir(os.path.join(project_folder, "node_modules"))


def find_package_json_files(base_path: str) -> List[str]:
    package_files = []
    for root, dirs, files in os.walk(base_path):
        if "package.json" in files:
            package_files.append(Path(root) / "package.json")
    return package_files


def get_all_node_modules_packages(project_folder: str) -> List[dict]:

    packages = []

    # Get all instances of the package.json file in the node_modules directory
    package_files = find_package_json_files(os.path.join(project_folder, "node_modules"))

    # Add all packages to the list
    for package_file in package_files:
        with open(package_file) as f:
            data = json.load(f)
            if "name" in data and "version" in data:
                name = data['name']
                version = data['version']

                # Make sure name is valid
                if name.startswith("<%="):
                    continue

                # Extract the name and version for each package
                packages.append({"name": name, "version": version})

    return packages


def package_lock_file_exists(project_folder: str) -> bool:
    return os.path.isfile(os.path.join(project_folder, "package-lock.json"))


def get_all_package_lock_packages(project_folder: str) -> List[dict]:
    # Get all packages in the package-lock.json file

    with open(os.path.join(project_folder, "package-lock.json")) as f:
        data = json.load(f)

        return [{"name": package.replace("node_modules/", ""), "version": data["packages"][package]["version"]} for package in data["packages"] if package != ""]


def package_json_file_exists(project_folder: str) -> bool:
    return os.path.isfile(os.path.join(project_folder, "package.json"))


def get_all_package_json_packages(project_folder: str) -> List[dict]:
    # Get all packages in the package.json file
    with open(os.path.join(project_folder, "package.json")) as f:
        data = json.load(f)

        deps = [{"name": dependency, "version": data["dependencies"][dependency]} for dependency in data["dependencies"]]
        deps.extend([{"name": dependency, "version": data["devDependencies"][dependency]} for dependency in data["devDependencies"]])

    return deps


def run_npm_dry_run(project_folder: str) -> str:
    return exec_command("npm install --dry-run --verbose", project_folder)


def get_all_npm_dry_run_packages(project_folder: str) -> List[dict]:
    # Get all packages from npm install --dry-run --verbose

    ret = []

    # Run the command
    result = run_npm_dry_run(project_folder)

    # Parse the output
    for line in result.split("\n"):
        # when the line starts with add
        if line.startswith("add"):
            # Extract the package name and version
            name = line.split(" ")[1]
            version = line.split(" ")[2]
            ret.append({"name": name, "version": version})

    # Extract the packages
    return ret


def yarn_lock_file_exists(project_folder: str) -> bool:
    return os.path.isfile(os.path.join(project_folder, "yarn.lock"))


def get_all_yarn_lock_packages(project_folder: str) -> List[dict]:
    # Get all packages in the yarn.lock file
    package_data = []
    file_path = os.path.join(project_folder, "yarn.lock")
    with open(file_path, 'r') as f:
        content = f.read()

    # Regex to match package entries and their version
    package_regex = r'^(?:"|)([^\s"][^"]*)(?:"|):\n\s+version\s+"([^"]+)"'

    # Find all matches in the yarn.lock content
    matches = re.finditer(package_regex, content, re.MULTILINE)

    for match in matches:
        package_names = match.group(1)  # Full package spec (can include multiple package names)
        version = match.group(2)       # Version

        # Split package names if there are multiple (comma-separated)
        for name in package_names.split(", "):

            # Remove the last @ and everything after it
            name = name.rsplit("@", 1)[0].strip()

            # Validate the name is not an alias
            alias_match = re.match(r"([^@]+)@npm:([^@]+)", name)

            if alias_match:
                _, name = alias_match.groups()

            package_data.append({"name": name, "version": version.strip()})

    return package_data


def run_yarn_install(project_folder: str) -> str:
    return exec_command("yarn install --check-files -non-interactive", project_folder)


def get_all_yarn_list_packages(project_folder: str) -> List[dict]:
    # Get all packages from yarn list

    ret = []

    # Run the command
    result = exec_command("yarn list --json --no-progress", project_folder)

    list_json = result.strip().split('\n')[-1]

    # Parse the output
    data = json.loads(list_json)

    # Extract the packages
    for package in data["data"]["trees"]:
        name_and_version = package["name"].rsplit("@", 1)
        ret.append({"name": name_and_version[0], "version": name_and_version[1]})

    return ret


# Global functions
def update_sbom_from_npm(ossbom: OSSBOM, project_folder: str) -> OSSBOM:

    # get all versions of a package in the node_modules directory
    if node_modules_directory_exists(project_folder):
        components = get_all_node_modules_packages(project_folder)
        ossbom.add_components([Component.create(name=component["name"], version=component["version"], env=DependencyEnv.PROD, type="npm", source="node_modules") for component in components])

    # get all packages in the package-lock.json file
    if package_lock_file_exists(project_folder):
        components = get_all_package_lock_packages(project_folder)
        ossbom.add_components([Component.create(name=component["name"], version=component["version"], env=DependencyEnv.PROD, type="npm", source="package-lock.json") for component in components])

    # npm install --dry-run --verbose
    components = get_all_npm_dry_run_packages(project_folder)
    ossbom.add_components([Component.create(name=component["name"], version=component["version"], env=DependencyEnv.PROD, type="npm", source="install") for component in components])

    return ossbom


def update_sbom_from_yarn(ossbom: OSSBOM, project_folder: str, run_install: bool = False) -> OSSBOM:

    if run_install:
        run_yarn_install(project_folder)

    # get all versions of a package in the node_modules directory
    if node_modules_directory_exists(project_folder):
        components = get_all_node_modules_packages(project_folder)
        ossbom.add_components([Component.create(name=component["name"], version=component["version"], env=DependencyEnv.PROD, type="npm", source="node_modules") for component in components])

    # get all packages in the package-lock.json file
    if yarn_lock_file_exists(project_folder):
        components = get_all_yarn_lock_packages(project_folder)
        ossbom.add_components([Component.create(name=component["name"], version=component["version"], env=DependencyEnv.PROD, type="npm", source="yarn.lock") for component in components])

    # get all packages in the package.json file
    #if package_json_file_exists(project_folder):
    #    packages.add_list(get_all_package_json_packages(project_folder), "package.json", DependencyEnv.PROD, type="npm")

    # yarn list
    components = get_all_yarn_list_packages(project_folder)
    ossbom.add_components([Component.create(name=component["name"], version=component["version"], env=DependencyEnv.PROD, type="npm", source="yarn list") for component in components])

    return ossbom



from ossprey.scan import scan
from ossbom.model.ossbom import OSSBOM
import pytest


@pytest.mark.parametrize("mode", ["python-requirements", "auto"])
def test_scan_py_success(mode):
    ret = scan("test/python_simple_math", mode=mode, local_scan=True)
    assert isinstance(ret, OSSBOM)
    assert [comp.name for comp in ret.get_components()] == ['numpy', 'requests']


def test_scan_py_success_pipenv():
    ret = scan("test/python_simple_math", mode="pipenv", local_scan=True)
    assert isinstance(ret, OSSBOM)

    result = [  
        'certifi',
        'charset-normalizer',
        'idna',
        'numpy',
        'requests',
        'simple_math',
        'urllib3'
    ]
    assert [comp.name for comp in ret.get_components()] == result


@pytest.mark.parametrize("mode", ["pipenv", "auto"])
def test_scan_poetry_success_pipenv(mode):
    ret = scan("test/poetry_simple_math", mode=mode, local_scan=True)
    assert isinstance(ret, OSSBOM)

    result = [  
        'certifi',
        'charset-normalizer',
        'idna',
        'numpy',
        'poetry-simple-math',
        'requests',
        'urllib3'
    ]
    assert [comp.name for comp in ret.get_components()] == result


@pytest.mark.parametrize("mode", ["npm", "auto"])
def test_scan_npm_success(mode):
    ret = scan("test/npm_simple_math", mode=mode, local_scan=True)
    assert isinstance(ret, OSSBOM)
    assert len(ret.get_components()) == 333


@pytest.mark.parametrize(["mode", "num_components"], [
    ("yarn", 323),
    ("auto", 324)
])
def test_scan_yarn_success(mode, num_components):
    ret = scan("test/yarn_simple_math", mode=mode, local_scan=True)
    assert isinstance(ret, OSSBOM)
    assert len(ret.get_components()) == num_components


def test_scan_failure():
    with pytest.raises(Exception) as excinfo:
        scan("test/python_simple_math_no_exist", mode="python-requirements", local_scan=True)
    assert "Package test/python_simple_math_no_exist does not exist" in str(excinfo.value)


def test_scan_invalid_mode():
    with pytest.raises(Exception) as excinfo:
        scan("test/python_simple_math", mode="invalid-mode", local_scan=True)
    assert "Invalid scanning method" in str(excinfo.value)




from ossprey.sbom_python import create_sbom_from_env, create_sbom_from_requirements
from ossprey.virtualenv import VirtualEnv


def test_get_sbom():
    sbom = create_sbom_from_env()

    assert sbom.format == 'OSSBOM'


def test_get_sbom_from_venv():

    venv = VirtualEnv()
    venv.enter()

    # Install a package
    venv.install_package('numpy')

    requirements_file = venv.create_requirements_file_from_env()

    # Get the SBOM
    sbom = create_sbom_from_requirements(requirements_file)

    assert sbom.format == 'OSSBOM'
    assert len(sbom.components) == 1
    assert any(map(lambda x: x.name == 'numpy', sbom.components.values()))


def test_get_sbom_from_venv_local_package():

    venv = VirtualEnv()
    venv.enter()

    # Install a package
    venv.install_package('test/python_simple_math')
  
    requirements_file = venv.create_requirements_file_from_env()

    # Get the SBOM
    sbom = create_sbom_from_requirements(requirements_file)

    assert sbom.format == 'OSSBOM'
    assert len(sbom.components) == 7
    assert any(map(lambda x: x.name == 'simple_math', sbom.components.values()))



import json
import logging
import os
import sys

from ossbom.converters.factory import SBOMConverterFactory

from ossprey.args import parse_arguments
from ossprey.github_actions_reporter import print_gh_action_errors
from ossprey.log import init_logging
from ossprey.sbom_python import update_sbom_from_requirements
from ossprey.sbom_javascript import update_sbom_from_npm, update_sbom_from_yarn
from ossprey.ossprey import Ossprey
from ossprey.virtualenv import VirtualEnv
from ossprey.environment import get_environment_details

from ossbom.model.ossbom import OSSBOM

logger = logging.getLogger(__name__)


def get_modes(directory):
    """
    Get the modes from the directory.
    :param directory: The directory to scan.
    :return: A list of modes.
    """

    # get all files in the directory
    files = [f for f in os.listdir(directory) if os.path.isfile(os.path.join(directory, f))]
    logging.debug(f"Files in directory: {files}")
    modes = []

    # Check for requirements.txt
    if "requirements.txt" in files:
        modes.append("python-requirements")

    poetry_files = [
        "poetry.lock",
        "pyproject.toml"
    ]
    if any(poetry_file in files for poetry_file in poetry_files):
        # TODO handle poetry better in the future
        modes.append("pipenv")

    npm_files = [
        "package-lock.json",
        "package.json",
        "node_modules"
    ]
    # Check for package.json
    if any(npm_file in files for npm_file in npm_files):
        modes.append("npm")

    # Check for yarn.lock
    if "yarn.lock" in files:
        modes.append("yarn")

    return modes


def scan(package_name, mode="auto", local_scan=False, url=None, api_key=None):

    if mode == "auto":
        logging.debug("Auto mode selected")
        
        # Check the folder for files that map to different package managers
        modes = get_modes(package_name)
        if len(modes) == 0:
            logging.error("No package manager found")
            raise Exception("No package manager found in the directory")
    else:
        modes = [mode]

    # If package location doesn't exist, raise an error
    if not os.path.exists(package_name):
        logging.error(f"Package {package_name} does not exist")
        raise Exception(f"Package {package_name} does not exist")

    sbom = OSSBOM()

    if "pipenv" in modes:
        venv = VirtualEnv()
        venv.enter()

        venv.install_package(package_name)
        requirements_file = venv.create_requirements_file_from_env()

        sbom = update_sbom_from_requirements(sbom, requirements_file)

        venv.exit()
    elif "python-requirements" in modes:
        sbom = update_sbom_from_requirements(sbom, package_name + "/requirements.txt")
    elif "npm" in modes:
        sbom = update_sbom_from_npm(sbom, package_name)
    elif "yarn" in modes:
        sbom = update_sbom_from_yarn(sbom, package_name)
    else:
        raise Exception("Invalid scanning method: " + str(modes))

    # Update sbom to contain the local environment
    env = get_environment_details(package_name)
    sbom.update_environment(env)

    logging.info(f"Scanning {len(sbom.get_components())}")

    if not local_scan:
        ossprey = Ossprey(url, api_key)

        # Compress to MINIBOM
        sbom = SBOMConverterFactory.to_minibom(sbom)

        sbom = ossprey.validate(sbom)
        if not sbom:
            raise Exception("Issue OSSPREY Service")

        # Convert to OSSBOM
        sbom = SBOMConverterFactory.from_minibom(sbom)

    logger.debug(json.dumps(sbom.to_dict(), indent=4))

    return sbom



import pytest
from unittest.mock import patch, mock_open
from ossprey.sbom_javascript import (
    exec_command, node_modules_directory_exists, find_package_json_files,
    get_all_node_modules_packages, package_lock_file_exists, get_all_package_lock_packages,
    package_json_file_exists, get_all_package_json_packages, run_npm_dry_run,
    get_all_npm_dry_run_packages, yarn_lock_file_exists, get_all_yarn_lock_packages,
    run_yarn_install, get_all_yarn_list_packages, update_sbom_from_npm, update_sbom_from_yarn
)

from ossbom.model.ossbom import OSSBOM


def test_exec_command():
    with patch("subprocess.run") as mock_run:
        mock_run.return_value.stdout = "output"
        assert exec_command("echo test") == "output"


def test_node_modules_directory_exists(tmp_path):
    (tmp_path / "node_modules").mkdir()
    assert node_modules_directory_exists(tmp_path) is True


def test_find_package_json_files(tmp_path):
    (tmp_path / "node_modules").mkdir()
    (tmp_path / "node_modules" / "package.json").write_text("{}")
    assert find_package_json_files(tmp_path) == [tmp_path / "node_modules" / "package.json"]


def test_get_all_node_modules_packages(tmp_path):
    (tmp_path / "node_modules").mkdir()
    (tmp_path / "node_modules" / "package.json").write_text('{"name": "testpkg", "version": "1.0.0"}')
    assert get_all_node_modules_packages(tmp_path) == [{"name": "testpkg", "version": "1.0.0"}]


def test_package_lock_file_exists(tmp_path):
    (tmp_path / "package-lock.json").write_text("{}")
    assert package_lock_file_exists(tmp_path) is True


def test_get_all_package_lock_packages(tmp_path):
    (tmp_path / "package-lock.json").write_text('{"packages": {"": {}, "node_modules/testpkg": {"version": "1.0.0"}}}')
    assert get_all_package_lock_packages(tmp_path) == [{"name": "testpkg", "version": "1.0.0"}]


def test_package_json_file_exists(tmp_path):
    (tmp_path / "package.json").write_text("{}")
    assert package_json_file_exists(tmp_path) is True


def test_get_all_package_json_packages(tmp_path):
    (tmp_path / "package.json").write_text('{"dependencies": {"testpkg": "1.0.0"}, "devDependencies": {"testpkg-dev": "1.0.0"}}')
    assert get_all_package_json_packages(tmp_path) == [{"name": "testpkg", "version": "1.0.0"}, {"name": "testpkg-dev", "version": "1.0.0"}]


def test_run_npm_dry_run():
    with patch("ossprey.sbom_javascript.exec_command") as mock_exec_command:
        mock_exec_command.return_value = "add testpkg 1.0.0\n"
        assert get_all_npm_dry_run_packages(".") == [{"name": "testpkg", "version": "1.0.0"}]


def test_yarn_lock_file_exists(tmp_path):
    (tmp_path / "yarn.lock").write_text("")
    assert yarn_lock_file_exists(tmp_path) is True


def test_get_all_yarn_lock_packages(tmp_path):
    (tmp_path / "yarn.lock").write_text('"testpkg@^1.0.0":\n  version "1.0.0"\n')
    assert get_all_yarn_lock_packages(tmp_path) == [{"name": "testpkg", "version": "1.0.0"}]


def test_run_yarn_install():
    with patch("ossprey.sbom_javascript.exec_command") as mock_exec_command:
        mock_exec_command.return_value = "output"
        assert run_yarn_install(".") == "output"


def test_get_all_yarn_list_packages():
    with patch("ossprey.sbom_javascript.exec_command") as mock_exec_command:
        mock_exec_command.return_value = '{"data": {"trees": [{"name": "testpkg@1.0.0"}]}}'
        assert get_all_yarn_list_packages(".") == [{"name": "testpkg", "version": "1.0.0"}]


def test_update_sbom_from_npm():
    with patch("ossprey.sbom_javascript.get_all_npm_dry_run_packages") as mock_get_all_npm_dry_run_packages:
        mock_get_all_npm_dry_run_packages.return_value = [{"name": "testpkg", "version": "1.0.0"}]
        sbom = OSSBOM()
        sbom = update_sbom_from_npm(sbom, ".")
        assert len(sbom.components) == 1

        # Get only entry in sbom.components and confirm it's name value is testpkg
        for component in sbom.components.values():
            assert component.name == "testpkg"


def test_update_sbom_from_yarn():
    with patch("ossprey.sbom_javascript.get_all_yarn_list_packages") as mock_get_all_yarn_list_packages:
        mock_get_all_yarn_list_packages.return_value = [{"name": "testpkg", "version": "1.0.0"}]
        sbom = OSSBOM()
        sbom = update_sbom_from_yarn(sbom, ".")
        assert len(sbom.components) == 1

        # Get only entry in sbom.components and confirm it's name value is testpkg
        for component in sbom.components.values():
            assert component.name == "testpkg"
